from typing import Literal, TypedDict


LiteralValueType = str | int | float | bool | list | None


ValueType = Literal["str", "int", "float", "bool", "list", "none"]


BinOps = Literal[
    "add",
    "sub",
    "and",
    "or",
    "eq",
    "ne",
    "lt",
    "gt",
    "le",
    "ge",
]


ArgObj = TypedDict('ArgObj', {
    "kind": Literal["arg"],
    "name": str,
    "readable": str,
})
KeyObj = TypedDict('KeyObj', {
    "kind": Literal["key"],
    "name": str,
    "readable": str,
})
VarObj = TypedDict('VarObj', {
    "kind": Literal["var"],
    "name": str,
})
IndexObj = TypedDict('IndexObj', {
    "kind": Literal["index"],
    "name": str,
})
RefIdObj = ArgObj | KeyObj | VarObj | IndexObj


LoadArg = TypedDict('LoadArg', {
    "kind": Literal["load_json_arg", "load_key_arg"],
    "index": int,
})
LiteralValObj = TypedDict('LiteralValObj', {
    "kind": Literal["val"],
    "value": LiteralValueType,
    "type": ValueType,
})
ConstantObj = TypedDict('ConstantObj', {
    "kind": Literal["constant"],
    "raw": str,
})
UnaryOpObj = TypedDict('UnaryOpObj', {
    "kind": Literal["unary"],
    "op": Literal["not"],
    "arg": 'ExprObj',
})
BinaryOpObj = TypedDict('BinaryOpObj', {
    "kind": Literal["binary"],
    "op": BinOps,
    "left": 'ExprObj',
    "right": 'ExprObj',
})
ArrayAtObj = TypedDict('ArrayAtObj', {
    "kind": Literal["array_at"],
    "var": RefIdObj,
    "index": 'ExprObj',
})
ArrayLengthObj = TypedDict('ArrayLengthObj', {
    "kind": Literal["array_len"],
    "var": RefIdObj,
})
CallObj = TypedDict('CallObj', {
    "kind": Literal["call"],
    "name": str,
    "args": 'list[ExprObj]',
    "no_adjust": bool,
})
ExprObj = (
    RefIdObj
    | LoadArg
    | LiteralValObj
    | ConstantObj
    | UnaryOpObj
    | BinaryOpObj
    | ArrayAtObj
    | ArrayLengthObj
    | CallObj
)



from collections.abc import Callable

from redipy.graph.expr import ExprObj, ValueType


class Expr:  # pylint: disable=too-few-public-methods
    def compile(self) -> ExprObj:
        raise NotImplementedError()

    def __add__(self, other: 'MixedType') -> 'Expr':
        return AddOp(self, other)

    def __sub__(self, other: 'MixedType') -> 'Expr':
        return SubOp(self, other)

    def eq_(self, other: 'MixedType') -> 'Expr':
        return EqOp(self, other)

    def ne_(self, other: 'MixedType') -> 'Expr':
        return NeOp(self, other)

    def lt_(self, other: 'MixedType') -> 'Expr':
        return LtOp(self, other)

    def le_(self, other: 'MixedType') -> 'Expr':
        return LeOp(self, other)

    def gt_(self, other: 'MixedType') -> 'Expr':
        return GtOp(self, other)

    def ge_(self, other: 'MixedType') -> 'Expr':
        return GeOp(self, other)

    def not_(self) -> 'Expr':
        return NotOp(self)

    def or_(self, other: 'MixedType') -> 'Expr':
        return OrOp(self, other)

    def and_(self, other: 'MixedType') -> 'Expr':
        return AndOp(self, other)


class ExprHelper(Expr):  # pylint: disable=too-few-public-methods
    def __init__(self, expr_fn: Callable[[], ExprObj]) -> None:
        super().__init__()
        self._expr_fn = expr_fn

    def compile(self) -> ExprObj:
        return self._expr_fn()


class Constant(Expr):
    def __init__(self, raw: str) -> None:
        super().__init__()
        self._raw = raw

    def compile(self) -> ExprObj:
        return {
            "kind": "constant",
            "raw": self._raw,
        }


JSONType = str | int | float | list | dict | None
LiteralType = str | int | float | bool | list | None
MixedType = LiteralType | Expr


class LiteralOp(Expr):
    def __init__(self, value: LiteralType) -> None:
        super().__init__()
        self._value = value
        self._type = self.compute_type(value)

    @staticmethod
    def compute_type(value: LiteralType) -> ValueType:
        if value is None:
            return "none"
        if isinstance(value, bool):
            return "bool"
        if isinstance(value, int):
            return "int"
        if isinstance(value, float):
            return "float"
        if isinstance(value, str):
            return "str"
        if isinstance(value, list):
            return "list"
        raise ValueError(f"unknown type for: {value}")

    def compile(self) -> ExprObj:
        return {
            "kind": "val",
            "value": self._value,
            "type": self._type,
        }


def lit_helper(value: MixedType) -> Expr:
    if isinstance(value, Expr):
        return value
    return LiteralOp(value)


class NotOp(Expr):
    def __init__(self, expr: MixedType) -> None:
        super().__init__()
        self._expr = lit_helper(expr)

    def compile(self) -> ExprObj:
        return {
            "kind": "unary",
            "op": "not",
            "arg": self._expr.compile(),
        }


class Op(Expr):
    def __init__(self, lhs: MixedType, rhs: MixedType) -> None:
        super().__init__()
        self._lhs = lit_helper(lhs)
        self._rhs = lit_helper(rhs)

    def get_left(self) -> Expr:
        return self._lhs

    def get_right(self) -> Expr:
        return self._rhs

    def compile(self) -> ExprObj:
        raise NotImplementedError()


class AndOp(Op):
    def compile(self) -> ExprObj:
        return {
            "kind": "binary",
            "op": "and",
            "left": self.get_left().compile(),
            "right": self.get_right().compile(),
        }


class OrOp(Op):
    def compile(self) -> ExprObj:
        return {
            "kind": "binary",
            "op": "or",
            "left": self.get_left().compile(),
            "right": self.get_right().compile(),
        }


class AddOp(Op):
    def compile(self) -> ExprObj:
        return {
            "kind": "binary",
            "op": "add",
            "left": self.get_left().compile(),
            "right": self.get_right().compile(),
        }


class SubOp(Op):
    def compile(self) -> ExprObj:
        return {
            "kind": "binary",
            "op": "sub",
            "left": self.get_left().compile(),
            "right": self.get_right().compile(),
        }


class LtOp(Op):
    def compile(self) -> ExprObj:
        return {
            "kind": "binary",
            "op": "lt",
            "left": self.get_left().compile(),
            "right": self.get_right().compile(),
        }


class LeOp(Op):
    def compile(self) -> ExprObj:
        return {
            "kind": "binary",
            "op": "le",
            "left": self.get_left().compile(),
            "right": self.get_right().compile(),
        }


class GtOp(Op):
    def compile(self) -> ExprObj:
        return {
            "kind": "binary",
            "op": "gt",
            "left": self.get_left().compile(),
            "right": self.get_right().compile(),
        }


class GeOp(Op):
    def compile(self) -> ExprObj:
        return {
            "kind": "binary",
            "op": "ge",
            "left": self.get_left().compile(),
            "right": self.get_right().compile(),
        }


class EqOp(Op):
    def compile(self) -> ExprObj:
        return {
            "kind": "binary",
            "op": "eq",
            "left": self.get_left().compile(),
            "right": self.get_right().compile(),
        }


class NeOp(Op):
    def compile(self) -> ExprObj:
        return {
            "kind": "binary",
            "op": "ne",
            "left": self.get_left().compile(),
            "right": self.get_right().compile(),
        }



from typing import overload

from redipy.backend.backend import ExecFunction
from redipy.symbolic.seq import FnContext


class RedisAPI:
    def register_script(self, ctx: FnContext) -> ExecFunction:
        raise NotImplementedError()

    def set(self, key: str, value: str) -> str:
        raise NotImplementedError()

    def get(self, key: str) -> str | None:
        raise NotImplementedError()

    def lpush(self, key: str, *values: str) -> int:
        raise NotImplementedError()

    def rpush(self, key: str, *values: str) -> int:
        raise NotImplementedError()

    @overload
    def lpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def lpop(
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def lpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        raise NotImplementedError()

    @overload
    def rpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def rpop(
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def rpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        raise NotImplementedError()

    def llen(self, key: str) -> int:
        raise NotImplementedError()

    def zadd(self, key: str, mapping: dict[str, float]) -> int:
        raise NotImplementedError()

    def zpop_max(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        raise NotImplementedError()

    def zpop_min(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        raise NotImplementedError()

    def zcard(self, key: str) -> int:
        raise NotImplementedError()



from collections.abc import Callable

from redipy.graph.cmd import AssignmentObj, CommandObj
from redipy.graph.expr import ArgObj, ExprObj, RefIdObj, VarObj
from redipy.symbolic.expr import Expr, ExprHelper, lit_helper, MixedType


class Compilable:  # pylint: disable=too-few-public-methods
    def compile(self) -> CommandObj:
        raise NotImplementedError()


class CmdHelper(Compilable):  # pylint: disable=too-few-public-methods
    def __init__(self, stmt_fn: Callable[[], CommandObj]) -> None:
        self._stmt_fn = stmt_fn

    def compile(self) -> CommandObj:
        return self._stmt_fn()


class Variable(Expr):
    def __init__(self) -> None:
        self._index: int | None = None

    def len_(self) -> 'Expr':
        return ArrayLen(self)

    def __getitem__(self, other: 'MixedType') -> 'Expr':
        return ArrayAt(self, other)

    def set_index(self, index: int) -> None:
        self._index = index

    def get_index(self) -> int:
        assert self._index is not None
        return self._index

    def prefix(self) -> str:
        raise NotImplementedError()

    def get_declare(self) -> AssignmentObj:
        return {
            "kind": "declare",
            "assign": self.get_ref(),
            "value": self.get_declare_rhs().compile(),
        }

    def get_declare_rhs(self) -> Expr:
        raise NotImplementedError()

    def get_ref(self) -> RefIdObj:
        raise NotImplementedError()

    def compile(self) -> RefIdObj:
        return self.get_ref()

    def assign(self, val: MixedType) -> CmdHelper:
        expr = lit_helper(val)
        return CmdHelper(
            lambda: {
                "kind": "assign",
                "assign": self.get_ref(),
                "value": expr.compile(),
            })

    def set_at(self, index: MixedType, val: MixedType) -> CmdHelper:
        ix = lit_helper(index)
        expr = lit_helper(val)
        return CmdHelper(
            lambda: {
                "kind": "assign_at",
                "assign": self.get_ref(),
                "index": ix.compile(),
                "value": expr.compile(),
            })


class JSONArg(Variable):
    def __init__(self, name: str) -> None:
        super().__init__()
        self._name = name

    def get_declare_rhs(self) -> Expr:
        return ExprHelper(
            lambda: {
                "kind": "load_json_arg",
                "index": self.get_index(),
            })

    def prefix(self) -> str:
        return "arg"

    def get_arg_ref(self) -> ArgObj:
        return {
            "kind": "arg",
            "name": f"{self.prefix()}_{self.get_index()}",
            "readable": self._name,
        }

    def get_ref(self) -> RefIdObj:
        return self.get_arg_ref()


class LocalVariable(Variable):
    def __init__(self, init: MixedType) -> None:
        super().__init__()
        self._init = lit_helper(init)

    def get_declare_rhs(self) -> Expr:
        return self._init

    def prefix(self) -> str:
        return "var"

    def get_var_ref(self) -> VarObj:
        return {
            "kind": "var",
            "name": f"{self.prefix()}_{self.get_index()}",
        }

    def get_ref(self) -> RefIdObj:
        return self.get_var_ref()


class ArrayAt(Expr):
    def __init__(self, array: Variable, index: MixedType) -> None:
        super().__init__()
        self._array = array
        self._index = lit_helper(index)

    def compile(self) -> ExprObj:
        return {
            "kind": "array_at",
            "var": self._array.get_ref(),
            "index": self._index.compile(),
        }


class ArrayLen(Expr):
    def __init__(self, array: Variable) -> None:
        super().__init__()
        self._array = array

    def compile(self) -> ExprObj:
        return {
            "kind": "array_len",
            "var": self._array.get_ref(),
        }



from typing import Literal, TYPE_CHECKING, TypedDict

from redipy.graph.expr import ExprObj, IndexObj, RefIdObj, VarObj


if TYPE_CHECKING:
    from redipy.graph.seq import SequenceObj


AssignmentObj = TypedDict('AssignmentObj', {
    "kind": Literal["assign", "declare"],
    "assign": RefIdObj,
    "value": ExprObj,
})
AssignAtObj = TypedDict('AssignAtObj', {
    "kind": Literal["assign_at"],
    "assign": RefIdObj,
    "index": ExprObj,
    "value": ExprObj,
})
StmtObj = TypedDict('StmtObj', {
    "kind": Literal["stmt"],
    "expr": ExprObj,
})
BranchObj = TypedDict('BranchObj', {
    "kind": Literal["branch"],
    "condition": ExprObj,
    "then": 'SequenceObj',
    "else": 'SequenceObj',
})
ForLoopObj = TypedDict('ForLoopObj', {
    "kind": Literal["for"],
    "array": ExprObj,
    "index": IndexObj,
    "value": VarObj,
    "body": 'SequenceObj',
})
WhileObj = TypedDict('WhileObj', {
    "kind": Literal["while"],
    "condition": ExprObj,
    "body": 'SequenceObj',
})
ReturnObj = TypedDict('ReturnObj', {
    "kind": Literal["return"],
    "value": ExprObj | None,
})
CommandObj = (
    AssignmentObj
    | AssignAtObj
    | StmtObj
    | BranchObj
    | ForLoopObj
    | WhileObj
    | ReturnObj
)



import contextlib
import threading
import uuid
from collections.abc import Callable, Iterable, Iterator
from contextlib import AbstractContextManager
from typing import Any, overload, Protocol, TypedDict

from redis import Redis
from redis.commands.core import Script
from redis.exceptions import ResponseError

from redipy.backend.runtime import Runtime
from redipy.redis.lua import LuaBackend
from redipy.util import is_test


RedisConfig = TypedDict('RedisConfig', {
    "host": str,
    "port": int,
    "passwd": str,
    "prefix": str,
    "path": str,
})


class RedisFactory(Protocol):  # pylint: disable=too-few-public-methods
    def __call__(self, *, cfg: RedisConfig) -> Redis:
        ...


class RedisFunctionBytes(Protocol):  # pylint: disable=too-few-public-methods
    def __call__(
            self,
            *,
            keys: list[str],
            args: list[Any],
            client: Redis | None) -> bytes:
        ...


CONCURRENT_MODULE_CONN: int = 17


class RedisWrapper:
    def __init__(
            self,
            *,
            cfg: RedisConfig,
            redis_factory: RedisFactory | None = None,
            is_caching_enabled: bool = True) -> None:
        self._redis_factory: RedisFactory = (
            self._create_connection
            if redis_factory is None else redis_factory)
        self._cfg = cfg
        self._is_caching_enabled = is_caching_enabled
        self._service_conn: list[Redis | None] = \
            [None] * CONCURRENT_MODULE_CONN
        self._lock = threading.RLock()

    @staticmethod
    def get_connection_index() -> int:
        return threading.get_ident() % CONCURRENT_MODULE_CONN

    @staticmethod
    def _create_connection(*, cfg: RedisConfig) -> Redis:
        return Redis(
            host=cfg["host"],
            port=cfg["port"],
            db=0,
            password=cfg["passwd"],
            retry_on_timeout=True,
            health_check_interval=45,
            client_name=f"rc-{uuid.uuid4().hex}")

    def _get_redis_cached_conn(self) -> Redis:
        if not self._is_caching_enabled:
            return self._redis_factory(cfg=self._cfg)

        ix = self.get_connection_index()
        res = self._service_conn[ix]
        if res is None:
            with self._lock:
                res = self._service_conn[ix]
                if res is None:
                    res = self._redis_factory(cfg=self._cfg)
                    self._service_conn[ix] = res
        return res

    @contextlib.contextmanager
    def get_connection(self) -> Iterator[Redis]:
        try:
            yield self._get_redis_cached_conn()
        except Exception:
            self.reset()
            raise

    def reset(self) -> None:
        with self._lock:
            for ix, _ in enumerate(self._service_conn):
                self._service_conn[ix] = None


class RedisConnection(Runtime[list[str]]):
    def __init__(
            self,
            redis_module: str,
            *,
            cfg: RedisConfig,
            redis_factory: RedisFactory | None = None,
            is_caching_enabled: bool = True) -> None:
        super().__init__()
        self._conn = RedisWrapper(
            cfg=cfg,
            redis_factory=redis_factory,
            is_caching_enabled=is_caching_enabled)
        prefix_str = f"{cfg['prefix']}:" if cfg["prefix"] else ""
        module = f"{prefix_str}{redis_module}".rstrip(":")
        self._module = f"{module}:" if module else ""

    @classmethod
    def create_backend(cls) -> LuaBackend:
        return LuaBackend()

    def get_connection(self) -> AbstractContextManager[Redis]:
        return self._conn.get_connection()

    def get_dynamic_script(self, code: str) -> RedisFunctionBytes:
        if is_test():
            print(
                "Compiled script:\n-- SCRIPT START --\n"
                f"{code.rstrip()}\n-- SCRIPT END --")
        compute = Script(None, code.encode("utf-8"))
        context = 3

        def get_error(err_msg: str) -> tuple[str, list[str]] | None:
            ustr = "user_script:"
            ix = err_msg.find(ustr)
            if ix < 0:
                return None
            eix = err_msg.find(":", ix + len(ustr))
            if eix < 0:
                return None
            num = int(err_msg[ix + len(ustr):eix])
            rel_line = num

            new_msg = f"{err_msg[:ix + len(ustr)]}{rel_line}{err_msg[eix:]}"
            ctx = [""] * context + code.splitlines()
            return new_msg, ctx[num - 1:num + 2 * context]

        @contextlib.contextmanager
        def get_client(client: Redis | None) -> Iterator[Redis]:
            try:
                if client is None:
                    with self.get_connection() as res:
                        yield res
                else:
                    yield client
            except ResponseError as e:
                handle_err(e)
                raise e

        def handle_err(exc: ResponseError) -> None:
            if exc.args:
                msg = exc.args[0]
                res = get_error(msg)
                if res is not None:
                    ctx = "\n".join((
                        f"{'>' if ix == context else ' '} {line}"
                        for (ix, line) in enumerate(res[1])))
                    exc.add_note(
                        f"{res[0].rstrip()}\nCode:\n{code}\n\nContext:\n{ctx}")

        def execute_bytes_result(
                *,
                keys: list[str],
                args: list[bytes | str | int],
                client: Redis | None) -> bytes:
            with get_client(client) as inner:
                return compute(keys=keys, args=args, client=inner)

        return execute_bytes_result

    def get_prefix(self) -> str:
        return self._module

    def with_prefix(self, key: str) -> str:
        return f"{self.get_prefix()}{key}"

    def get_pubsub_key(self, key: str) -> str:
        return f"{self.get_prefix()}ps:{key}"

    def wait_for(
            self,
            key: str,
            predicate: Callable[[], bool],
            granularity: float = 30.0) -> None:
        if predicate():
            return
        with self.get_connection() as conn:
            with conn.pubsub() as psub:
                psub.subscribe(self.get_pubsub_key(key))
                try:
                    while not predicate():
                        psub.get_message(
                            ignore_subscribe_messages=True,
                            timeout=granularity)
                        while psub.get_message() is not None:  # flushing queue
                            pass
                finally:
                    psub.unsubscribe()

    def notify_all(self, key: str) -> None:
        with self.get_connection() as conn:
            conn.publish(self.get_pubsub_key(key), "notify")

    def ping(self) -> None:
        with self.get_connection() as conn:
            conn.ping()

    def flush_all(self) -> None:
        with self.get_connection() as conn:
            conn.flushall()

    def keys_count(self, prefix: str) -> int:
        full_prefix = f"{prefix}*"
        vals: set[bytes] = set()
        cursor = 0
        count = 10
        with self.get_connection() as conn:
            while True:
                cursor, res = conn.scan(cursor, full_prefix, count)
                vals.update(res)
                if cursor == 0:
                    break
                if count < 4000:
                    count = min(4000, count * 2)
        return len(vals)

    def keys_str(
            self, prefix: str, postfix: str | None = None) -> Iterable[str]:
        full_prefix = f"{prefix}*{'' if postfix is None else postfix}"
        vals: set[bytes] = set()
        cursor = 0
        count = 10
        with self.get_connection() as conn:
            while True:
                cursor, res = conn.scan(cursor, full_prefix, count)
                vals.update(res)
                if cursor == 0:
                    break
                if count < 1000:
                    count = int(min(1000, count * 1.2))
        return (val.decode("utf-8") for val in vals)

    def prefix_exists(
            self, prefix: str, postfix: str | None = None) -> bool:
        full_prefix = f"{prefix}*{'' if postfix is None else postfix}"
        cursor = 0
        count = 10
        with self.get_connection() as conn:
            while True:
                cursor, res = conn.scan(cursor, full_prefix, count)
                if res:
                    return True
                if cursor == 0:
                    return False
                if count < 1000:
                    count = int(min(1000, count * 1.2))

    def set(self, key: str, value: str) -> str:
        with self.get_connection() as conn:
            conn.set(self.with_prefix(key), value)
            return "OK"

    def get(self, key: str) -> str | None:
        with self.get_connection() as conn:
            res = conn.get(self.with_prefix(key))
            if res is None:
                return None
            return res.decode("utf-8")

    def lpush(self, key: str, *values: str) -> int:
        with self.get_connection() as conn:
            return conn.lpush(self.with_prefix(key), *values)

    def rpush(self, key: str, *values: str) -> int:
        with self.get_connection() as conn:
            return conn.rpush(self.with_prefix(key), *values)

    @overload
    def lpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def lpop(  # pylint: disable=signature-differs
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def lpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        with self.get_connection() as conn:
            res = conn.lpop(self.with_prefix(key), count)
            if res is None:
                return None
            if count is None:
                return res.decode("utf-8")
            return [val.decode("utf-8") for val in res]

    @overload
    def rpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def rpop(  # pylint: disable=signature-differs
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def rpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        with self.get_connection() as conn:
            res = conn.rpop(self.with_prefix(key), count)
            if res is None:
                return None
            if count is None:
                return res.decode("utf-8")
            return [val.decode("utf-8") for val in res]

    def llen(self, key: str) -> int:
        with self.get_connection() as conn:
            return conn.llen(self.with_prefix(key))

    def zadd(self, key: str, mapping: dict[str, float]) -> int:
        with self.get_connection() as conn:
            return int(
                conn.zadd(self.with_prefix(key), mapping))  # type: ignore

    def zpop_max(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        with self.get_connection() as conn:
            res = conn.zpopmax(self.with_prefix(key), count)
            return [
                (name.decode("utf-8"), float(score))
                for name, score in res
            ]

    def zpop_min(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        with self.get_connection() as conn:
            res = conn.zpopmin(self.with_prefix(key), count)
            return [
                (name.decode("utf-8"), float(score))
                for name, score in res
            ]

    def zcard(self, key: str) -> int:
        with self.get_connection() as conn:
            return int(conn.zcard(self.with_prefix(key)))



import datetime
import hashlib
import inspect
import json
import os
import re
import string
import threading
import uuid
from collections.abc import Callable, Iterable
from typing import Any, IO, TypeVar

import pytz


ET = TypeVar('ET')
CT = TypeVar('CT')
RT = TypeVar('RT')
VT = TypeVar('VT')


NL = "\n"


TEST_SALT_LOCK = threading.RLock()
TEST_SALT: dict[str, str] = {}


def is_test() -> bool:
    test_id = os.getenv("PYTEST_CURRENT_TEST")
    return test_id is not None


def get_test_salt() -> str | None:
    test_id = os.getenv("PYTEST_CURRENT_TEST")
    if test_id is None:
        return None
    res = TEST_SALT.get(test_id)
    if res is None:
        with TEST_SALT_LOCK:
            res = TEST_SALT.get(test_id)
            if res is None:
                res = f"salt:{uuid.uuid4().hex}"
                TEST_SALT[test_id] = res
    return res


def indent(text: str, amount: int = 0) -> list[str]:
    istr = " " * amount
    return [
        f"{istr}{line}" for line in text.splitlines()
    ]


def deindent(text: str) -> str:
    min_indent = None
    lines = text.rstrip().splitlines()
    if lines and not lines[0]:
        lines.pop(0)
    for line in lines:
        cur_indent = len(line) - len(line.lstrip())
        if min_indent is None or min_indent > cur_indent:
            min_indent = cur_indent
            if min_indent == 0:
                break
    return "".join((f"{line[min_indent:]}\n" for line in lines))


def lua_fmt(text: str) -> str:
    return deindent(text).replace("    ", "  ")


def code_fmt(lines: list[str]) -> str:
    return "".join(f"{line.rstrip()}\n" for line in lines)


def get_text_hash(text: str) -> str:
    blake = hashlib.blake2b(digest_size=32)
    blake.update(text.encode("utf-8"))
    return blake.hexdigest()


def text_hash_size() -> int:
    return 64


def get_short_hash(text: str) -> str:
    blake = hashlib.blake2b(digest_size=4)
    blake.update(text.encode("utf-8"))
    return blake.hexdigest()


def short_hash_size() -> int:
    return 8


BUFF_SIZE = 65536  # 64KiB


def get_file_hash(fname: str) -> str:
    blake = hashlib.blake2b(digest_size=32)
    with open(fname, "rb") as fin:
        while True:
            buff = fin.read(BUFF_SIZE)
            if not buff:
                break
            blake.update(buff)
    return blake.hexdigest()


def file_hash_size() -> int:
    return 64


def is_hex(text: str) -> bool:
    hex_digits = set(string.hexdigits)
    return all(char in hex_digits for char in text)


def only(arr: list[RT]) -> RT:
    if len(arr) != 1:
        raise ValueError(f"array must have exactly one element: {arr}")
    return arr[0]


# time units for logging request durations
ELAPSED_UNITS: list[tuple[int, str]] = [
    (1, "s"),
    (60, "m"),
    (60*60, "h"),
    (60*60*24, "d"),
]


def elapsed_time_string(elapsed: float) -> str:
    """Convert elapsed time into a readable string."""
    cur = ""
    for (conv, unit) in ELAPSED_UNITS:
        if elapsed / conv >= 1 or not cur:
            cur = f"{elapsed / conv:8.3f}{unit}"
        else:
            break
    return cur


def to_bool(value: bool | float | int | str) -> bool:
    value = f"{value}".lower()
    if value == "true":
        return True
    if value == "false":
        return False
    try:
        return bool(int(float(value)))
    except ValueError:
        pass
    raise ValueError(f"{value} cannot be interpreted as bool")


def to_list(value: Any) -> list[Any]:
    if not isinstance(value, list):
        raise ValueError(f"{value} is not a list")
    return value


def is_int(value: Any) -> bool:
    try:
        int(value)
        return True
    except ValueError:
        return False


def is_float(value: Any) -> bool:
    try:
        float(value)
        return True
    except (ValueError, TypeError):
        return False


def is_json(value: str) -> bool:
    try:
        json.loads(value)
    except json.JSONDecodeError:
        return False
    return True


def report_json_error(err: json.JSONDecodeError) -> None:
    raise ValueError(
        f"JSON parse error ({err.lineno}:{err.colno}): "
        f"{repr(err.doc)}") from err


def json_maybe_read(data: str) -> Any | None:
    try:
        return json.loads(data)
    except json.JSONDecodeError:
        return None


def json_load(fin: IO[str]) -> Any:
    try:
        return json.load(fin)
    except json.JSONDecodeError as e:
        report_json_error(e)
        raise e


def json_dump(obj: Any, fout: IO[str]) -> None:
    print(json_pretty(obj), file=fout)


def json_pretty(obj: Any) -> str:
    return json.dumps(obj, sort_keys=True, indent=2)


def json_compact(obj: Any) -> bytes:
    return json.dumps(
        obj,
        sort_keys=True,
        indent=None,
        separators=(",", ":")).encode("utf-8")


def json_read(data: bytes) -> Any:
    try:
        return json.loads(data.decode("utf-8"))
    except json.JSONDecodeError as e:
        report_json_error(e)
        raise e


def read_jsonl(fin: IO[str]) -> Iterable[Any]:
    for line in fin:
        line = line.rstrip()
        if not line:
            continue
        try:
            yield json.loads(line)
        except json.JSONDecodeError as e:
            report_json_error(e)
            raise e


def from_timestamp(timestamp: float) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(timestamp, pytz.utc)


def to_timestamp(time: datetime.datetime) -> float:
    return time.timestamp()


def now_ts() -> datetime.datetime:
    return datetime.datetime.now(pytz.utc)


def get_function_info(*, clazz: type) -> tuple[str, int, str]:
    stack = inspect.stack()

    def get_method(cur_clazz: type) -> tuple[str, int, str] | None:
        class_filename = inspect.getfile(cur_clazz)
        for level in stack:
            if os.path.samefile(level.filename, class_filename):
                return level.filename, level.lineno, level.function
        return None

    queue = [clazz]
    while queue:
        cur = queue.pop(0)
        res = get_method(cur)
        if res is not None:
            return res
        queue.extend(cur.__bases__)
    frame = stack[1]
    return frame.filename, frame.lineno, frame.function


def get_relative_function_info(
        depth: int) -> tuple[str, int, str, dict[str, Any]]:
    depth += 1
    stack = inspect.stack()
    if depth >= len(stack):
        return "unknown", -1, "unknown", {}
    frame = stack[depth]
    return frame.filename, frame.lineno, frame.function, frame.frame.f_locals


def identity(obj: RT) -> RT:
    return obj


NUMBER_PATTERN = re.compile(r"\d+")


def extract_list(
        arr: Iterable[str],
        prefix: str | None = None,
        postfix: str | None = None) -> Iterable[tuple[str, str]]:
    if not arr:
        yield from []
        return

    for elem in arr:
        text = elem
        if prefix is not None:
            if not text.startswith(prefix):
                continue
            text = text[len(prefix):]
        if postfix is not None:
            if not text.endswith(postfix):
                continue
            text = text[:-len(postfix)]
        yield (elem, text)


def extract_number(
        arr: Iterable[str],
        prefix: str | None = None,
        postfix: str | None = None) -> Iterable[tuple[str, int]]:

    def get_num(text: str) -> int | None:
        match = re.search(NUMBER_PATTERN, text)
        if match is None:
            return None
        try:
            return int(match.group())
        except ValueError:
            return None

    for elem, text in extract_list(arr, prefix=prefix, postfix=postfix):
        num = get_num(text)
        if num is None:
            continue
        yield elem, num


def highest_number(
        arr: Iterable[str],
        prefix: str | None = None,
        postfix: str | None = None) -> tuple[str, int] | None:
    res = None
    res_num = 0
    for elem, num in extract_number(arr, prefix=prefix, postfix=postfix):
        if res is None or num > res_num:
            res = elem
            res_num = num
    return None if res is None else (res, res_num)


def retain_some(
        arr: Iterable[VT],
        count: int,
        *,
        key: Callable[[VT], Any],
        reverse: bool = False,
        keep_last: bool = True) -> tuple[list[VT], list[VT]]:
    res: list[VT] = []
    to_delete: list[VT] = []
    if keep_last:
        for elem in arr:
            if len(res) <= count:
                res.append(elem)
                continue
            res.sort(key=key, reverse=reverse)
            to_delete.extend(res[:-count])
            res = res[-count:]
            res.append(elem)
    else:
        for elem in arr:
            res.append(elem)
            if len(res) < count:
                continue
            res.sort(key=key, reverse=reverse)
            to_delete.extend(res[:-count])
            res = res[-count:]
    res.sort(key=key, reverse=reverse)
    return res, to_delete


def python_module() -> str:
    stack = inspect.stack()
    module = inspect.getmodule(stack[1][0])
    if module is None:
        raise ValueError("module not found")
    res = module.__name__
    if res != "__main__":
        return res
    package = module.__package__
    if package is None:
        package = ""
    mfname = module.__file__
    if mfname is None:
        return package
    fname = os.path.basename(mfname)
    fname = fname.removesuffix(".py")
    if fname in ("__init__", "__main__"):
        return package
    return f"{package}.{fname}"


def parent_python_module(p_module: str) -> str:
    dot_ix = p_module.rfind(".")
    if dot_ix < 0:
        return ""
    return p_module[:dot_ix]


def check_pid_exists(pid: int) -> bool:
    try:
        os.kill(pid, 0)
        return True
    except OSError:
        return False


def ideal_thread_count() -> int:
    res = os.cpu_count()
    if res is None:
        return 4
    return res


def escape(text: str, subs: dict[str, str]) -> str:
    text = text.replace("\\", "\\\\")
    for key, repl in subs.items():
        text = text.replace(key, f"\\{repl}")
    return text


def unescape(text: str, subs: dict[str, str]) -> str:
    res: list[str] = []
    in_escape = False
    for c in text:
        if in_escape:
            in_escape = False
            if c == "\\":
                res.append("\\")
                continue
            done = False
            for key, repl in subs.items():
                if c == key:
                    res.append(repl)
                    done = True
                    break
            if done:
                continue
        if c == "\\":
            in_escape = True
            continue
        res.append(c)
    return "".join(res)



from typing import Generic, Protocol, TYPE_CHECKING, TypeVar

from redipy.graph.cmd import CommandObj
from redipy.graph.expr import ExprObj
from redipy.graph.seq import SequenceObj
from redipy.symbolic.expr import JSONType


if TYPE_CHECKING:
    from redipy.backend.runtime import Runtime


T = TypeVar('T')
ET = TypeVar('ET')
CC = TypeVar('CC')
EC = TypeVar('EC')
R = TypeVar('R', bound='Runtime')


class ExecFunction(Protocol):  # pylint: disable=too-few-public-methods
    def __call__(
            self,
            keys: dict[str, str],
            args: dict[str, JSONType]) -> JSONType:
        raise NotImplementedError()


class Backend(Generic[T, ET, CC, EC, R]):
    def translate(self, seq: SequenceObj) -> T:
        ctx = self.create_command_context()
        res = self.compile_sequence(ctx, seq)
        return self.finish(ctx, res)

    def create_command_context(self) -> CC:
        raise NotImplementedError()

    def finish(self, ctx: CC, script: T) -> T:
        raise NotImplementedError()

    def compile_sequence(self, ctx: CC, seq: SequenceObj) -> T:
        raise NotImplementedError()

    def compile_command(self, ctx: CC, cmd: CommandObj) -> T:
        raise NotImplementedError()

    def compile_expr(self, ctx: EC, expr: ExprObj) -> ET:
        raise NotImplementedError()

    def create_executable(
            self,
            code: T,
            runtime: R) -> ExecFunction:
        raise NotImplementedError()



import json
from test.util import get_test_config

from redipy.memory.local import LocalBackend
from redipy.memory.rt import LocalRuntime
from redipy.redis.conn import RedisConnection
from redipy.redis.lua import LuaBackend
from redipy.symbolic.expr import JSONType
from redipy.symbolic.fun import ToNum
from redipy.symbolic.rvar import RedisVar
from redipy.symbolic.seq import FnContext
from redipy.util import code_fmt, lua_fmt


LUA_SCRIPT = lua_fmt("""
--[[ KEYV
k
]]
--[[ ARGV
a
]]
local arg_0 = cjson.decode(ARGV[1])  -- a
local key_0 = (KEYS[1])  -- k
if (tonumber((redis.call("get", key_0) or 0)) <= arg_0) then
    redis.call("set", key_0, arg_0)
end
local var_0 = (redis.call("get", key_0) or nil)
if (var_0 ~= nil) then
    var_0 = tonumber(var_0)
end
return cjson.encode(var_0)
""")


RUN_TESTS: list[tuple[str, JSONType, JSONType]] = [
    ("foo", 1, 1),
    ("foo", 3, 3),
    ("foo", 2, 3),
    ("bar", 5, 5),
    ("bar", 2, 5),
    ("foo", 4, 4),
    ("bar", 5, 5),
    ("foo", 8, 8),
    ("foo", 9, 9),
    ("neg", -1, None),
    ("neg", 0, 0),
    ("neg", 1, 1),
    ("neg", 0, 1),
    ("bar", 2, 5),
    ("pos", 3, 5),
    ("bar", 7, 7),
    ("pos", 6, 6),
]


def test_rvar() -> None:
    ctx = FnContext()
    a = ctx.add_arg("a")
    k = RedisVar(ctx.add_key("k"))
    b_then, _ = ctx.if_(ToNum(k.get(no_adjust=True).or_(0)).le_(a))
    b_then.add(k.set(a))
    res = ctx.add_local(k.get())
    r_then, _ = ctx.if_(res.ne_(None))
    r_then.add(res.assign(ToNum(res)))
    ctx.set_return_value(res)

    compiled = ctx.compile()
    print(json.dumps(compiled, indent=2, sort_keys=True))

    lcl = LocalBackend()
    lcl_code = lcl.translate(compiled)
    lrt = LocalRuntime()
    run_lcl = lcl.create_executable(lcl_code, lrt)

    lrt.set("pos", "5")
    for (k_in, a_in, expect_out) in RUN_TESTS:
        is_out = run_lcl({"k": k_in}, {"a": a_in})
        assert is_out == expect_out
    assert lrt.get("foo") == "9"
    assert lrt.get("bar") == "7"
    assert lrt.get("neg") == "1"
    assert lrt.get("pos") == "6"

    lua = LuaBackend()
    lua_code = lua.translate(compiled)
    assert code_fmt(lua_code) == LUA_SCRIPT

    rrt = RedisConnection("test_rvar", cfg=get_test_config())
    run_redis = lua.create_executable(lua_code, rrt)

    rrt.set("pos", "5")
    for (k_in, a_in, expect_out) in RUN_TESTS:
        is_out = run_redis({"k": k_in}, {"a": a_in})
        assert is_out == expect_out
    assert rrt.get("foo") == "9"
    assert rrt.get("bar") == "7"
    assert rrt.get("neg") == "1"
    assert rrt.get("pos") == "6"



from redipy.symbolic.expr import Expr, MixedType
from redipy.symbolic.fun import KeyVariable, RedisFn


class RedisSortedSet:
    def __init__(self, key: KeyVariable) -> None:
        self._key = key

    def add(self, score: MixedType, value: MixedType) -> Expr:
        # TODO add all arguments
        return RedisFn("zadd", self._key, score, value)

    def pop_max(self, count: MixedType = None) -> Expr:
        if count is None:
            return RedisFn("zpopmax", self._key)
        return RedisFn("zpopmax", self._key, count)

    def pop_min(self, count: MixedType = None) -> Expr:
        if count is None:
            return RedisFn("zpopmin", self._key)
        return RedisFn("zpopmin", self._key, count)

    def card(self) -> Expr:
        return RedisFn("zcard", self._key)



from redipy.symbolic.expr import Expr, MixedType
from redipy.symbolic.fun import KeyVariable, RedisFn


class RedisList:
    def __init__(self, key: KeyVariable) -> None:
        self._key = key

    def lpush(self, *values: MixedType) -> Expr:
        return RedisFn("lpush", self._key, *values)

    def rpush(self, *values: MixedType) -> Expr:
        return RedisFn("rpush", self._key, *values)

    def lpop(
            self, count: MixedType = None, *, no_adjust: bool = False) -> Expr:
        if count is None:
            return RedisFn("lpop", self._key, no_adjust=no_adjust)
        return RedisFn("lpop", self._key, count, no_adjust=no_adjust)

    def rpop(
            self, count: MixedType = None, *, no_adjust: bool = False) -> Expr:
        if count is None:
            return RedisFn("rpop", self._key, no_adjust=no_adjust)
        return RedisFn("rpop", self._key, count, no_adjust=no_adjust)

    def llen(self) -> Expr:
        return RedisFn("llen", self._key)



import redipy.script  # pylint: disable=unused-import  # noqa
from redipy.backend.backend import ExecFunction
from redipy.backend.runtime import Runtime
from redipy.main import Redis
from redipy.memory.rt import LocalRuntime
from redipy.redis.conn import RedisConfig, RedisConnection, RedisFactory


__all__ = [
    "ExecFunction",
    "LocalRuntime",
    "Redis",
    "RedisConfig",
    "RedisConnection",
    "RedisFactory",
    "Runtime",
    "script",
]



import collections
import json
from typing import cast, overload

from redipy.backend.runtime import Runtime
from redipy.memory.local import Cmd, LocalBackend
from redipy.symbolic.expr import JSONType


CONST: dict[str, JSONType] = {
    "redis.LOG_DEBUG": "DEBUG",
    "redis.LOG_VERBOSE": "VERBOSE",
    "redis.LOG_NOTICE": "NOTICE",
    "redis.LOG_WARNING": "WARNING",
}


class LocalRuntime(Runtime[Cmd]):
    def __init__(self) -> None:
        super().__init__()
        self._vals: dict[str, str] = {}
        self._queues: collections.defaultdict[str, collections.deque[str]] = \
            collections.defaultdict(collections.deque)
        # TODO: to be replaced by something more efficient later
        self._zorder: collections.defaultdict[str, list[str]] = \
            collections.defaultdict(list)
        self._zscores: collections.defaultdict[str, dict[str, float]] = \
            collections.defaultdict(dict)

    @classmethod
    def create_backend(cls) -> LocalBackend:
        return LocalBackend()

    @staticmethod
    def require_argc(
            args: list[JSONType],
            count: int,
            *,
            at_least: bool = False,
            at_most: int | None = None) -> None:
        argc = len(args)
        if argc == count:
            return
        if at_most is not None and argc <= at_most:
            return
        if at_least and argc > count:
            return
        raise ValueError(
            "incorrect number of arguments need "
            f"{'at least' if at_least else 'exactly'} {count} got {argc}")

    def redis_fn(
            self, name: str, args: list[JSONType]) -> JSONType:
        key = f"{args[0]}"
        if name == "set":
            self.require_argc(args, 2)
            return self.set(key, f"{args[1]}")
        if name == "setnx":
            # TODO move up and just redirect
            self.require_argc(args, 2)
            if self._vals.get(key) is not None:
                return 0
            self._vals[key] = f"{args[1]}"
            return 1
        if name == "get":
            self.require_argc(args, 1)
            return self.get(key)
        if name == "lpush":
            self.require_argc(args, 2, at_least=True)
            return self.lpush(key, *(f"{arg}" for arg in args[1:]))
        if name == "rpush":
            self.require_argc(args, 2, at_least=True)
            return self.rpush(key, *(f"{arg}" for arg in args[1:]))
        if name == "lpop":
            self.require_argc(args, 1, at_most=2)
            return self.lpop(
                key, None if len(args) < 2 else int(cast(int, args[1])))
        if name == "rpop":
            self.require_argc(args, 1, at_most=2)
            return self.rpop(
                key, None if len(args) < 2 else int(cast(int, args[1])))
        if name == "llen":
            self.require_argc(args, 1)
            return self.llen(key)
        if name == "zadd":
            self.require_argc(args, 2)
            return self.zadd(key, {f"{args[1]}": float(cast(float, args[2]))})
        if name == "zpopmax":
            self.require_argc(args, 1, at_most=2)
            return cast(list | None, self.zpop_max(
                key, 1 if len(args) < 2 else int(cast(int, args[1]))))
        if name == "zpopmin":
            self.require_argc(args, 1, at_most=2)
            return cast(list | None, self.zpop_min(
                key, 1 if len(args) < 2 else int(cast(int, args[1]))))
        if name == "zcard":
            self.require_argc(args, 1)
            return self.zcard(key)
        raise ValueError(f"unknow redis function {name}")

    def call_fn(
            self, name: str, args: list[JSONType]) -> JSONType:
        if name == "redis.call":
            self.require_argc(args, 2, at_least=True)
            return self.redis_fn(f"{args[0]}", args[1:])
        if name == "string.find":
            self.require_argc(args, 2, at_most=3)
            found_ix = f"{args[0]}".find(
                f"{args[1]}",
                int(cast(int, args[2])) if len(args) > 2 else None)
            return None if found_ix < 0 else found_ix
        if name == "cjson.decode":
            self.require_argc(args, 1)
            return json.loads(f"{args[0]}")
        if name == "cjson.encode":
            self.require_argc(args, 1)
            return json.dumps(
                f"{args[0]}",
                sort_keys=True,
                indent=None,
                separators=(",", ":"))
        if name == "tonumber":
            self.require_argc(args, 1)
            val = cast(str, args[0])
            try:
                return int(val)
            except (ValueError, TypeError):
                return float(val)
        if name == "tostring":
            self.require_argc(args, 1)
            return f"{args[0]}"
        if name == "redis.log":
            self.require_argc(args, 2)
            print(f"{args[0]}: {args[1]}")
            return None
        raise ValueError(f"unknown function {name}")

    def get_constant(self, raw: str) -> JSONType:
        return CONST[raw]

    def set(self, key: str, value: str) -> str:
        with self.lock():
            # TODO implement rest of arguments https://redis.io/commands/set/
            self._vals[key] = value
            return "OK"

    def get(self, key: str) -> str | None:
        with self.lock():
            return self._vals.get(key)

    def lpush(self, key: str, *values: str) -> int:
        with self.lock():
            queue = self._queues[key]
            queue.extendleft(values)
            return len(queue)

    def rpush(self, key: str, *values: str) -> int:
        with self.lock():
            queue = self._queues[key]
            queue.extend(values)
            return len(queue)

    @overload
    def lpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def lpop(  # pylint: disable=signature-differs
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def lpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        with self.lock():
            queue = self._queues[key]
            if not queue:
                return None
            if count is None:
                return queue.popleft()
            popc = count
            res = []
            while popc > 0 and queue:
                res.append(queue.popleft())
                popc -= 1
            return res if res else None

    @overload
    def rpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def rpop(  # pylint: disable=signature-differs
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def rpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        with self.lock():
            queue = self._queues[key]
            if not queue:
                return None
            if count is None:
                return queue.pop()
            popc = count
            res = []
            while popc > 0 and queue:
                res.append(queue.pop())
                popc -= 1
            return res if res else None

    def llen(self, key: str) -> int:
        with self.lock():
            mqueue = self._queues.get(key)
            return 0 if mqueue is None else len(mqueue)

    def zadd(self, key: str, mapping: dict[str, float]) -> int:
        with self.lock():
            count = 0
            zscores = self._zscores[key]
            zorder = self._zorder[key]
            for name, score in mapping.items():
                if name not in zscores:
                    zorder.append(name)
                    count += 1
                zscores[name] = score
            zorder.sort(key=lambda k: (zscores[k], k))
            return count

    def zpop_max(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        with self.lock():
            zscores = self._zscores[key]
            zorder = self._zorder[key]
            res = []
            remain = 1 if count is None else count
            while remain > 0 and zorder:
                name = zorder.pop()
                score = zscores.pop(name)
                res.append((name, score))
                remain -= 1
            return res

    def zpop_min(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        with self.lock():
            zscores = self._zscores[key]
            zorder = self._zorder[key]
            res = []
            remain = 1 if count is None else count
            while remain > 0 and zorder:
                name = zorder.pop(0)
                score = zscores.pop(name)
                res.append((name, score))
                remain -= 1
            return res

    def zcard(self, key: str) -> int:
        with self.lock():
            return len(self._zorder[key])

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}["
            f"vals={self._vals},"
            f"queues={dict(self._queues)},"
            f"zorder={dict(self._zorder)},"
            f"zscores={dict(self._zscores)}]")

    def __repr__(self) -> str:
        return self.__str__()



import contextlib
import threading
from collections.abc import Callable, Iterator
from typing import Any, Generic, Self, TypeVar

from redipy.api import RedisAPI
from redipy.backend.backend import Backend, ExecFunction
from redipy.graph.seq import SequenceObj
from redipy.symbolic.seq import FnContext


T = TypeVar('T')


class Runtime(Generic[T], RedisAPI):
    def __init__(self) -> None:
        self._backend: Backend[T, Any, Any, Any, Self] | None = None
        self._compile_hook: Callable[[SequenceObj], None] | None = None
        self._code_hook: Callable[[T], None] | None = None
        self._lock = threading.RLock()

    @contextlib.contextmanager
    def lock(self) -> Iterator[None]:
        with self._lock:
            yield

    def set_compile_hook(self, hook: Callable[[SequenceObj], None]) -> None:
        self._compile_hook = hook

    def set_code_hook(self, hook: Callable[[T], None]) -> None:
        self._code_hook = hook

    def register_script(self, ctx: FnContext) -> ExecFunction:
        compiled = ctx.compile()
        if self._compile_hook is not None:
            self._compile_hook(compiled)

        backend = self.get_backend()
        code = backend.translate(compiled)
        if self._code_hook is not None:
            self._code_hook(code)
        return backend.create_executable(code, self)

    def get_backend(self) -> Backend[T, Any, Any, Any, Self]:
        if self._backend is None:
            with self.lock():
                if self._backend is None:
                    self._backend = self.create_backend()
        return self._backend

    @classmethod
    def create_backend(cls) -> Backend[T, Any, Any, Any, Self]:
        raise NotImplementedError()



from typing import Literal, overload

from redipy.api import RedisAPI
from redipy.backend.backend import ExecFunction
from redipy.backend.runtime import Runtime
from redipy.memory.rt import LocalRuntime
from redipy.redis.conn import RedisConfig, RedisConnection, RedisFactory
from redipy.symbolic.seq import FnContext


class Redis(RedisAPI):
    def __init__(
            self,
            backend: Literal["memory", "redis", "custom"],
            *,
            cfg: RedisConfig | None = None,
            host: str | None = None,
            port: int | None = None,
            passwd: str | None = None,
            redis_module: str | None = None,
            prefix: str | None = None,
            path: str | None = None,
            is_caching_enabled: bool = True,
            redis_factory: RedisFactory | None = None,
            rt: Runtime | None = None,
            ) -> None:
        if backend == "custom":
            if rt is None:
                raise ValueError("rt must not be None for custom backend")
        elif backend == "memory":
            rt = LocalRuntime()
        elif backend == "redis":
            if cfg is None:
                cfg = {
                    "host": "localhost" if host is None else host,
                    "port": 6379 if port is None else port,
                    "passwd": "" if passwd is None else passwd,
                    "prefix": "" if prefix is None else prefix,
                    "path": "." if path is None else path,
                }
            rt = RedisConnection(
                "" if redis_module is None else redis_module,
                cfg=cfg,
                redis_factory=redis_factory,
                is_caching_enabled=is_caching_enabled)
        else:
            raise ValueError(f"unknown backend {backend}")
        self._rt: Runtime = rt

    def register_script(self, ctx: FnContext) -> ExecFunction:
        return self._rt.register_script(ctx)

    def set(self, key: str, value: str) -> str:
        return self._rt.set(key, value)

    def get(self, key: str) -> str | None:
        return self._rt.get(key)

    def lpush(self, key: str, *values: str) -> int:
        return self._rt.lpush(key, *values)

    def rpush(self, key: str, *values: str) -> int:
        return self._rt.rpush(key, *values)

    @overload
    def lpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def lpop(  # pylint: disable=signature-differs
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def lpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        return self._rt.lpop(key, count)

    @overload
    def rpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def rpop(  # pylint: disable=signature-differs
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def rpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        return self._rt.rpop(key, count)

    def llen(self, key: str) -> int:
        return self._rt.llen(key)

    def zadd(self, key: str, mapping: dict[str, float]) -> int:
        return self._rt.zadd(key, mapping)

    def zpop_max(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        return self._rt.zpop_max(key, count)

    def zpop_min(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        return self._rt.zpop_min(key, count)

    def zcard(self, key: str) -> int:
        return self._rt.zcard(key)



from redipy.symbolic.expr import Expr, MixedType
from redipy.symbolic.fun import KeyVariable, RedisFn


class RedisVar:
    def __init__(self, key: KeyVariable) -> None:
        self._key = key

    def set(self, value: MixedType) -> Expr:
        # TODO add all arguments
        return RedisFn("set", self._key, value)

    def get(self, *, no_adjust: bool = False) -> Expr:
        return RedisFn("get", self._key, no_adjust=no_adjust)



import json
from collections.abc import Iterable
from typing import Literal, TYPE_CHECKING

from redipy.backend.backend import Backend, ExecFunction
from redipy.graph.cmd import CommandObj
from redipy.graph.expr import BinOps, CallObj, ExprObj
from redipy.graph.seq import SequenceObj
from redipy.symbolic.expr import JSONType
from redipy.util import code_fmt, indent, json_compact, lua_fmt


if TYPE_CHECKING:
    from redipy.redis.conn import RedisConnection


HELPER_PKG = "redipy"


HELPER_FNS: dict[str, tuple[str, str]] = {
    "pairlist": (
        "arr",
        lua_fmt(r"""
            local res = {}
            local key = nil
            for ix, value in ipairs(arr) do
                if ix % 2 == 1 then
                    key = value
                else
                    res[#res + 1] = {key, value}
                end
            end
            return res
        """),
    ),
    "nil_or_index": (
        "val",
        lua_fmt(r"""
            if val ~= nil then
                val = val - 1
            end
            return val
        """),
    ),
}


class LuaFnHook:
    def __init__(self) -> None:
        self._helpers: set[str] = set()

    def build_helpers(self) -> list[str]:
        res = []
        prefix = f"{HELPER_PKG}."
        for helper in sorted(self._helpers):
            short_name = helper.removeprefix(prefix)
            args, body = HELPER_FNS[short_name]
            res.append(f"function {helper} ({args})")
            res.extend(indent(body, 2))
            res.append("end")
        return res

    def adjust_function(self, expr: CallObj) -> ExprObj:
        expr["no_adjust"] = True
        name = expr["name"]
        args = expr["args"]
        if name == "redis.call":
            r_name_obj = args[0]
            if r_name_obj["kind"] == "val" and r_name_obj["type"] == "str":
                return self.adjust_redis_fn(
                    expr, f"{r_name_obj['value']}", args[1:])
        if name == "string.find":
            return {
                "kind": "call",
                "name": f"{HELPER_PKG}.nil_or_index",
                "args": [expr],
                "no_adjust": False,
            }
        if name.startswith(f"{HELPER_PKG}."):
            self._helpers.add(name)
            return expr
        return expr

    def adjust_redis_fn(
            self,
            expr: CallObj,
            name: str,
            _args: list[ExprObj]) -> ExprObj:
        if name in ["get", "lpop", "rpop"]:
            return {
                "kind": "binary",
                "op": "or",
                "left": expr,
                "right": {
                    "kind": "val",
                    "type": "none",
                    "value": None,
                },
            }
        if name in ["zpopmax", "zpopmin"]:
            return {
                "kind": "call",
                "name": f"{HELPER_PKG}.pairlist",
                "args": [expr],
                "no_adjust": False,
            }
        return expr


def indent_str(code: Iterable[str], add_indent: int) -> list[str]:
    ind = " " * add_indent
    return [f"{ind}{exe}" for exe in code]


KEYV_HOOK = "--[[ KEYV"
ARGV_HOOK = "--[[ ARGV"
HOOK_END = "]]"


class LuaBackend(
        Backend[list[str], str, LuaFnHook, LuaFnHook, 'RedisConnection']):
    def create_command_context(self) -> LuaFnHook:
        return LuaFnHook()

    def finish(self, ctx: LuaFnHook, script: list[str]) -> list[str]:
        res = []
        helpers = ctx.build_helpers()
        if helpers:
            res.append("-- HELPERS START --")
            res.append(f"local {HELPER_PKG} = {{}}")
            res.extend(helpers)
            res.append("-- HELPERS END --")
        res.extend(script)
        return res

    def compile_sequence(self, ctx: LuaFnHook, seq: SequenceObj) -> list[str]:
        res = []
        if seq["kind"] == "script":
            key_order = seq["keyv"]
            arg_order = seq["argv"]
            res.append(KEYV_HOOK)
            for key in key_order:
                res.append(key)
            res.append(HOOK_END)
            res.append(ARGV_HOOK)
            for arg in arg_order:
                res.append(arg)
            res.append(HOOK_END)

        res.extend(indent_str(
            (
                exe
                for cmd in seq["cmds"]
                for exe in self.compile_command(ctx, cmd)
            ),
            0 if seq["kind"] == "script" else 2))
        return res

    def compile_command(self, ctx: LuaFnHook, cmd: CommandObj) -> list[str]:
        # FIXME: add debug context in comments
        if cmd["kind"] == "assign" or cmd["kind"] == "declare":
            is_declare = cmd["kind"] == "declare"
            decl = "local " if is_declare else ""
            rhs = self.compile_expr(ctx, cmd["value"])
            assign_obj = cmd["assign"]
            lcl_name = assign_obj["name"]
            if assign_obj["kind"] == "arg" or assign_obj["kind"] == "key":
                ext_name = f"  -- {assign_obj['readable']}"
            else:
                ext_name = ""
            return [f"{decl}{lcl_name} = {rhs}{ext_name}"]
        if cmd["kind"] == "assign_at":
            arr_ix = self.compile_expr(ctx, cmd["index"])
            rhs = self.compile_expr(ctx, cmd["value"])
            assign_obj = cmd["assign"]
            if assign_obj["kind"] == "var":
                lcl_name = assign_obj["name"]
                return [f"{lcl_name}[{arr_ix} + 1] = {rhs}"]
            raise ValueError(
                f"cannot assign to position of {cmd['assign']['kind']}")
        if cmd["kind"] == "stmt":
            stmt = self.compile_expr(ctx, cmd["expr"])
            return [stmt]
        if cmd["kind"] == "branch":
            res = [f"if {self.compile_expr(ctx, cmd['condition'])} then"]
            res.extend(self.compile_sequence(ctx, cmd["then"]))
            elses = self.compile_sequence(ctx, cmd["else"])
            if elses:
                res.append("else")
                res.extend(elses)
            res.append("end")
            return res
        if cmd["kind"] == "for":
            assert cmd["index"]["kind"] == "index"
            assert cmd["value"]["kind"] == "var"
            res = [(
                f"for {cmd['index']['name']}, {cmd['value']['name']} "
                f"in ipairs({self.compile_expr(ctx, cmd['array'])}) do"
            )]
            res.extend(self.compile_sequence(ctx, cmd["body"]))
            res.append("end")
            return res
        if cmd["kind"] == "while":
            res = [f"while {self.compile_expr(ctx, cmd['condition'])} do"]
            res.extend(self.compile_sequence(ctx, cmd["body"]))
            res.append("end")
            return res
        if cmd["kind"] == "return":
            if cmd["value"] is None:
                return []
            return [
                f"return cjson.encode({self.compile_expr(ctx, cmd['value'])})",
            ]
        raise ValueError(f"unknown kind {cmd['kind']} for command {cmd}")

    def compile_expr(self, ctx: LuaFnHook, expr: ExprObj) -> str:
        if expr["kind"] == "var":
            return expr["name"]
        if expr["kind"] == "arg":
            return expr["name"]
        if expr["kind"] == "key":
            return expr["name"]
        if expr["kind"] == "index":
            return f"({expr['name']} - 1)"
        if expr["kind"] == "load_json_arg":
            return f"cjson.decode(ARGV[{expr['index'] + 1}])"
        if expr["kind"] == "load_key_arg":
            return f"(KEYS[{expr['index'] + 1}])"
        if expr["kind"] == "val":
            val_type = expr["type"]
            value = expr["value"]
            if val_type == "bool":
                return f"{value}".lower()
            if val_type in ("int", "float"):
                return f"{value}"
            if val_type == "str":
                res = f"{value}"
                res = res.replace("\"", "\\\"").replace("\n", "\\n")
                return f"\"{res}\""
            if val_type == "list":
                res = json_compact(value).decode("utf-8")
                res = res.replace("\"", "\\\"").replace("\n", "\\n")
                return f"cjson.decode(\"{res}\")"
            if val_type == "none":
                return "nil"
            raise ValueError(f"unknown value type {val_type} for {expr}")
        if expr["kind"] == "unary":
            if expr["op"] == "not":
                return f"(not {self.compile_expr(ctx, expr['arg'])})"
            raise ValueError(f"unknown op {expr['op']} for {expr}")
        if expr["kind"] == "binary":
            ops: dict[BinOps, str] = {
                "and": "and",
                "or": "or",
                "add": "+",
                "sub": "-",
                "lt": "<",
                "le": "<=",
                "gt": ">",
                "ge": ">=",
                "eq": "==",
                "ne": "~=",
            }
            op = ops.get(expr["op"])
            if op is not None:
                return (
                    f"({self.compile_expr(ctx, expr['left'])} "
                    f"{op} "
                    f"{self.compile_expr(ctx, expr['right'])})")
            raise ValueError(f"unknown op {expr['op']} for {expr}")
        if expr["kind"] == "constant":
            return f"{expr['raw']}"
        if expr["kind"] == "array_at":
            return (
                f"{self.compile_expr(ctx, expr['var'])}"
                f"[{self.compile_expr(ctx, expr['index'])} + 1]")
        if expr["kind"] == "array_len":
            return f"#{self.compile_expr(ctx, expr['var'])}"
        if expr["kind"] == "call":
            if not expr["no_adjust"]:
                adj_expr = ctx.adjust_function(expr)
                return self.compile_expr(ctx, adj_expr)
            argstr = ", ".join(
                self.compile_expr(ctx, arg) for arg in expr["args"])
            return f"{expr['name']}({argstr})"
        raise ValueError(f"unknown kind {expr['kind']} for expression {expr}")

    def create_executable(
            self,
            code: list[str],
            runtime: 'RedisConnection') -> ExecFunction:
        key_order = []
        arg_order = []
        mode: Literal["none", "keyv", "argv"] = "none"
        for line in code:
            ln = line.strip()
            if ln == HOOK_END:
                mode = "none"
                continue
            if ln == KEYV_HOOK:
                mode = "keyv"
                continue
            if ln == ARGV_HOOK:
                mode = "argv"
                continue
            if mode == "keyv":
                key_order.append(ln)
                continue
            if mode == "argv":
                arg_order.append(ln)
                continue
        compute = runtime.get_dynamic_script(code_fmt(code))

        def exec_code(
                keys: dict[str, str],
                args: dict[str, JSONType]) -> JSONType:
            with runtime.get_connection() as client:
                res = compute(
                    keys=[runtime.with_prefix(keys[key]) for key in key_order],
                    args=[json_compact(args[arg]) for arg in arg_order],
                    client=client)
            if res is None:
                return None
            # NOTE: it is impossible to distinguish between {} and [] in lua
            if res == br"{}":
                return None
            return json.loads(res)

        return exec_code



from redipy.graph.cmd import CommandObj, StmtObj
from redipy.graph.expr import IndexObj, RefIdObj, VarObj
from redipy.graph.seq import SequenceObj
from redipy.symbolic.core import (
    CmdHelper,
    Compilable,
    JSONArg,
    LocalVariable,
    Variable,
)
from redipy.symbolic.expr import Expr, lit_helper, MixedType
from redipy.symbolic.fun import KeyVariable


class Sequence:
    def __init__(self, ctx: 'FnContext') -> None:
        self._ctx = ctx
        self._seq: list[Compilable] = []

    def add(
            self,
            term: Compilable | Expr) -> None:
        if isinstance(term, Compilable):
            self._seq.append(term)
        else:
            stmt: StmtObj = {
                "kind": "stmt",
                "expr": term.compile(),
            }
            self._seq.append(CmdHelper(lambda: stmt))

    def is_empty(self) -> bool:
        return not self._seq

    def get_cmds(self) -> list[CommandObj]:
        return [stmt.compile() for stmt in self._seq]

    def for_(self, array: Expr) -> tuple['Sequence', Variable, Variable]:
        loop = ForLoop(self._ctx, array)
        self.add(loop)
        return loop.get_loop(), loop.get_index(), loop.get_value()

    def if_(self, condition: MixedType) -> tuple['Sequence', 'Sequence']:
        branch = Branch(self._ctx, condition)
        self.add(branch)
        return branch.get_success(), branch.get_failure()

    def while_(self, condition: MixedType) -> 'Sequence':
        loop = WhileLoop(self._ctx, condition)
        self.add(loop)
        return loop.get_loop()


class Seq(Sequence):
    def compile(self) -> SequenceObj:
        return {
            "kind": "seq",
            "cmds": self.get_cmds(),
        }


class FnContext(Sequence):
    def __init__(self) -> None:
        super().__init__(self)
        self._args: list[tuple[str, JSONArg]] = []
        self._keys: list[tuple[str, KeyVariable]] = []
        self._anames: set[str] = set()
        self._knames: set[str] = set()
        self._locals: list[LocalVariable] = []
        self._loops: int = 0

    def add_arg(self, name: str) -> JSONArg:
        if name in self._anames:
            raise ValueError(f"ambiguous arg name: {name}")
        arg = JSONArg(name)
        arg.set_index(len(self._args))
        self._args.append((name, arg))
        self._anames.add(name)
        self.add(CmdHelper(arg.get_declare))
        return arg

    def add_key(self, name: str) -> KeyVariable:
        if name in self._knames:
            raise ValueError(f"ambiguous key name: {name}")
        key = KeyVariable(name)
        key.set_index(len(self._keys))
        self._keys.append((name, key))
        self._knames.add(name)
        self.add(CmdHelper(key.get_declare))
        return key

    def add_local(self, init: MixedType) -> LocalVariable:
        local = LocalVariable(init)
        local.set_index(len(self._locals))
        self._locals.append(local)
        self.add(CmdHelper(local.get_declare))
        return local

    def add_loop(self) -> int:
        loop = self._loops
        self._loops += 1
        return loop

    def set_return_value(self, value: MixedType) -> None:
        expr = lit_helper(value)
        self.add(CmdHelper(lambda: {
            "kind": "return",
            "value": expr.compile(),
        }))

    def compile(self) -> SequenceObj:
        return {
            "kind": "script",
            "cmds": self.get_cmds(),
            "argv": [arg for arg, _ in self._args],
            "keyv": [key for key, _ in self._keys],
        }


class Branch(Compilable):
    def __init__(self, ctx: FnContext, condition: MixedType) -> None:
        self._condition = lit_helper(condition)
        self._success = Seq(ctx)
        self._failure = Seq(ctx)

    def get_success(self) -> Sequence:
        return self._success

    def get_failure(self) -> Sequence:
        return self._failure

    def compile(self) -> CommandObj:
        return {
            "kind": "branch",
            "condition": self._condition.compile(),
            "then": self._success.compile(),
            "else": self._failure.compile(),
        }


class IndexVariable(Variable):
    def get_declare_rhs(self) -> Expr:
        raise RuntimeError("must be used in for loop")

    def prefix(self) -> str:
        return "ix"

    def get_index_ref(self) -> IndexObj:
        return {
            "kind": "index",
            "name": f"{self.prefix()}_{self.get_index()}",
        }

    def get_ref(self) -> RefIdObj:
        return self.get_index_ref()


class ValueVariable(Variable):
    def get_declare_rhs(self) -> Expr:
        raise RuntimeError("must be used in for loop")

    def prefix(self) -> str:
        return "val"

    def get_var_ref(self) -> VarObj:
        return {
            "kind": "var",
            "name": f"{self.prefix()}_{self.get_index()}",
        }

    def get_ref(self) -> RefIdObj:
        return self.get_var_ref()


class ForLoop(Compilable):
    def __init__(self, ctx: FnContext, array: Expr) -> None:
        loop_ix = ctx.add_loop()
        self._ix = IndexVariable()
        self._ix.set_index(loop_ix)
        self._val = ValueVariable()
        self._val.set_index(loop_ix)
        self._loop = Seq(ctx)
        self._array = array

    def get_index(self) -> Variable:
        return self._ix

    def get_value(self) -> Variable:
        return self._val

    def get_loop(self) -> Sequence:
        return self._loop

    def compile(self) -> CommandObj:
        return {
            "kind": "for",
            "array": self._array.compile(),
            "index": self._ix.get_index_ref(),
            "value": self._val.get_var_ref(),
            "body": self._loop.compile(),
        }


class WhileLoop(Compilable):
    def __init__(self, ctx: FnContext, condition: MixedType) -> None:
        self._condition = lit_helper(condition)
        self._loop = Seq(ctx)

    def get_loop(self) -> Sequence:
        return self._loop

    def compile(self) -> CommandObj:
        return {
            "kind": "while",
            "condition": self._condition.compile(),
            "body": self._loop.compile(),
        }



import json
from test.util import get_test_config

from redipy.memory.local import LocalBackend
from redipy.memory.rt import LocalRuntime
from redipy.redis.conn import RedisConnection
from redipy.redis.lua import LuaBackend
from redipy.symbolic.expr import JSONType
from redipy.symbolic.seq import FnContext
from redipy.util import code_fmt, lua_fmt


LUA_SCRIPT = lua_fmt("""
--[[ KEYV
]]
--[[ ARGV
a
b
]]
local arg_0 = cjson.decode(ARGV[1])  -- a
local arg_1 = cjson.decode(ARGV[2])  -- b
local var_0 = 5
local var_1 = 0.0
if ((arg_0 + arg_1) >= 10) then
    var_0 = (arg_0 - arg_1)
    var_1 = 2.5
else
    var_1 = 7.5
end
return cjson.encode((var_0 + var_1))
""")


RUN_TESTS: list[tuple[JSONType, JSONType, JSONType]] = [
    (2, 4, 12.5),
    (3, 7, -1.5),
    (13, 2, 13.5),
]


def test_simple() -> None:
    ctx = FnContext()
    a = ctx.add_arg("a")
    b = ctx.add_arg("b")
    c = ctx.add_local(5)
    d = ctx.add_local(0.0)
    b_then, b_else = ctx.if_((a + b).ge_(10))
    b_then.add(c.assign(a - b))
    b_then.add(d.assign(2.5))
    b_else.add(d.assign(7.5))
    ctx.set_return_value(c + d)

    compiled = ctx.compile()
    print(json.dumps(compiled, indent=2, sort_keys=True))

    lcl = LocalBackend()
    lcl_code = lcl.translate(compiled)
    lrt = LocalRuntime()
    run_lcl = lcl.create_executable(lcl_code, lrt)

    for (a_in, b_in, out) in RUN_TESTS:
        res = run_lcl({}, {"a": a_in, "b": b_in})
        assert res == out

    lua = LuaBackend()
    lua_code = lua.translate(compiled)
    assert code_fmt(lua_code) == LUA_SCRIPT

    conn = RedisConnection("test_simple", cfg=get_test_config())
    run_redis = lua.create_executable(lua_code, conn)

    for (a_in, b_in, out) in RUN_TESTS:
        res = run_redis({}, {"a": a_in, "b": b_in})
        assert res == out



from typing import Literal

from redipy.graph.expr import ExprObj, KeyObj, RefIdObj
from redipy.symbolic.core import Variable
from redipy.symbolic.expr import (
    Constant,
    Expr,
    ExprHelper,
    lit_helper,
    MixedType,
)


class KeyVariable(Variable):
    def __init__(self, name: str) -> None:
        super().__init__()
        self._name = name

    def get_declare_rhs(self) -> Expr:
        return ExprHelper(
            lambda: {
                "kind": "load_key_arg",
                "index": self.get_index(),
            })

    def prefix(self) -> str:
        return "key"

    def get_key_ref(self) -> KeyObj:
        return {
            "kind": "key",
            "name": f"{self.prefix()}_{self.get_index()}",
            "readable": self._name,
        }

    def get_ref(self) -> RefIdObj:
        return self.get_key_ref()


class CallFn(Expr):
    def __init__(
            self,
            fname: str,
            *args: MixedType,
            no_adjust: bool = False) -> None:
        self._fname = fname
        self._args: list[Expr] = [lit_helper(arg) for arg in args]
        self._no_adjust = no_adjust

    def compile(self) -> ExprObj:
        return {
            "kind": "call",
            "name": self._fname,
            "args": [arg.compile() for arg in self._args],
            "no_adjust": self._no_adjust,
        }


class FindFn(CallFn):
    def __init__(
            self,
            haystack: MixedType,
            needle: MixedType,
            start_ix: MixedType = None) -> None:
        super().__init__(
            "string.find",
            haystack,
            needle,
            *([] if start_ix is None else [start_ix]))


class FromJSON(CallFn):
    def __init__(self, arg: MixedType) -> None:
        super().__init__("cjson.decode", arg)


class ToJSON(CallFn):
    def __init__(self, arg: MixedType) -> None:
        super().__init__("cjson.encode", arg)


class ToNum(CallFn):
    def __init__(self, arg: MixedType) -> None:
        super().__init__("tonumber", arg)


class ToStr(CallFn):
    def __init__(self, arg: MixedType) -> None:
        super().__init__("tostr", arg)


class RedisFn(CallFn):
    def __init__(
            self,
            redis_fn: str,
            key: KeyVariable,
            *args: MixedType,
            no_adjust: bool = False) -> None:
        super().__init__(
            "redis.call", redis_fn, key, *args, no_adjust=no_adjust)


LogLevel = Literal["debug", "verbose", "notice", "warning"]


class LogFn(CallFn):
    def __init__(
            self,
            level: LogLevel,
            message: MixedType) -> None:
        super().__init__("redis.log", self.convert_level(level), message)

    @staticmethod
    def convert_level(
            level: LogLevel) -> Expr:
        level_map: dict[LogLevel, str] = {
            "debug": "redis.LOG_DEBUG",
            "verbose": "redis.LOG_VERBOSE",
            "notice": "redis.LOG_NOTICE",
            "warning": "redis.LOG_WARNING",
        }
        raw = level_map.get(level)
        if raw is None:
            raise ValueError(f"could not determine debug level: {level}")
        return Constant(raw)



# pylint: disable=unused-import
from redipy.symbolic.fun import (  # noqa
    FindFn,
    FromJSON,
    LogFn,
    LogLevel,
    ToJSON,
    ToNum,
    ToStr,
)
from redipy.symbolic.rlist import RedisList  # noqa
from redipy.symbolic.rvar import RedisVar  # noqa
from redipy.symbolic.rzset import RedisSortedSet  # noqa
from redipy.symbolic.seq import FnContext  # noqa



from test.util import get_setup, run_code

import pytest

from redipy.backend.backend import ExecFunction
from redipy.symbolic.fun import ToNum
from redipy.symbolic.rlist import RedisList
from redipy.symbolic.seq import FnContext
from redipy.util import lua_fmt


LUA_SCRIPT = lua_fmt("""
--[[ KEYV
inp
left
right
]]
--[[ ARGV
cmp
]]
local arg_0 = cjson.decode(ARGV[1])  -- cmp
local key_0 = (KEYS[1])  -- inp
local key_1 = (KEYS[2])  -- left
local key_2 = (KEYS[3])  -- right
local var_0 = (redis.call("lpop", key_0) or nil)
while (var_0 ~= nil) do
    if (tonumber(var_0) < arg_0) then
        redis.call("lpush", key_1, var_0)
    else
        redis.call("rpush", key_2, var_0)
    end
    var_0 = (redis.call("lpop", key_0) or nil)
end
if (redis.call("llen", key_1) > redis.call("llen", key_2)) then
    redis.call("lpush", key_2, (redis.call("rpop", key_1) or nil))
end
return cjson.encode(nil)
""")


RUN_TESTS: list[tuple[tuple[str, str, str, int], tuple[int, int]]] = [
    # a: 5, 2, 3, 7, 9, 8, 8, 1, 4
    (("a", "b", "c", 5), (4, 5)),
    # a:
    # b: 4, 1, 3, 2
    # c: 5, 7, 9, 8, 8
    # d: 1, 2, 3, 4, 5, 6, 7, 8, 9
    (("d", "b", "e", 7), (9, 4)),
    # d:
    # b: 6, 5, 4, 3, 2, 1, 4, 1, 3
    # e: 2, 7, 8, 9
    (("b", "a", "d", 3), (3, 6)),
    # b:
    # a: 1, 1, 2
    # d: 6, 5, 4, 3, 4, 3
]


@pytest.mark.parametrize("rt_lua", [False, True])
def test_rlist(rt_lua: bool) -> None:
    rt = get_setup("test_rlist", rt_lua, lua_script=LUA_SCRIPT)

    ctx = FnContext()
    cmp = ctx.add_arg("cmp")
    inp = RedisList(ctx.add_key("inp"))
    left = RedisList(ctx.add_key("left"))
    right = RedisList(ctx.add_key("right"))
    cur = ctx.add_local(inp.lpop())
    loop = ctx.while_(cur.ne_(None))
    b_then, b_else = loop.if_(ToNum(cur).lt_(cmp))
    b_then.add(left.lpush(cur))
    b_else.add(right.rpush(cur))
    loop.add(cur.assign(inp.lpop()))
    r_then, _ = ctx.if_(left.llen().gt_(right.llen()))
    r_then.add(right.lpush(left.rpop()))
    ctx.set_return_value(None)

    rt.rpush("a", "5", "2", "3", "7", "9", "8", "8", "1", "4")
    rt.lpush("d", "9", "8", "7", "6", "5", "4", "3", "2", "1")

    def tester(
            runner: ExecFunction,
            vals: tuple[str, str, str, int]) -> tuple[int, int]:
        k_inp, k_left, k_right, a_cmp = vals
        res = runner(
            keys={
                "inp": k_inp,
                "left": k_left,
                "right": k_right,
            },
            args={
                "cmp": a_cmp,
            })
        assert res is None
        assert rt.llen(k_inp) == 0
        return (rt.llen(k_left), rt.llen(k_right))

    run_code(rt, ctx, tests=RUN_TESTS, tester=tester)
    # a: 1, 1, 2
    # b:
    # c: 5, 7, 9, 8, 8
    # d: 6, 5, 4, 3, 4, 3
    # e: 2, 7, 8, 9
    assert rt.rpop("a", 3) == ["2", "1", "1"]
    assert rt.rpop("b", 3) is None
    assert rt.lpop("b", 3) is None
    assert rt.lpop("c", 3) == ["5", "7", "9"]
    assert rt.lpop("c", 3) == ["8", "8"]
    assert rt.rpop("d", 7) == ["3", "4", "3", "4", "5", "6"]
    assert rt.lpop("e", 4) == ["2", "7", "8", "9"]
